@@ -17,6 +17,7 @@
 #include "js/ContextOptions.h"        // JS::ContextOptionsRef
 #include "js/ErrorReport.h"           // JSErrorBase
 #include "js/friend/ErrorMessages.h"  // js::GetErrorMessage, JSMSG_*
+#include "js/friend/StackLimits.h"    // js::AutoCheckRecursionLimit
 #include "js/Modules.h"  // JS::FinishDynamicModuleImport, JS::{G,S}etModuleResolveHook, JS::Get{ModulePrivate,ModuleScript,RequestedModule{s,Specifier,SourcePos}}, JS::SetModule{DynamicImport,Metadata}Hook
 #include "js/PropertyAndElement.h"  // JS_DefineProperty, JS_GetElement
 #include "js/SourceText.h"
@@ -76,7 +77,8 @@ NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(ModuleLoaderBase)
 NS_INTERFACE_MAP_END
 
 NS_IMPL_CYCLE_COLLECTION(ModuleLoaderBase, mFetchingModules, mFetchedModules,
-                         mDynamicImportRequests, mGlobalObject, mLoader)
+                         mDynamicImportRequests, mGlobalObject, mOverriddenBy,
+                         mLoader)
 
 NS_IMPL_CYCLE_COLLECTING_ADDREF(ModuleLoaderBase)
 NS_IMPL_CYCLE_COLLECTING_RELEASE(ModuleLoaderBase)
@@ -286,6 +288,10 @@ bool ModuleLoaderBase::HostPopulateImportMeta(
   return true;
 }
 
+static bool ModuleTypeAllowed(JS::ModuleType aModuleType) {
+  return aModuleType != JS::ModuleType::Unknown;
+}
+
 // static
 bool ModuleLoaderBase::HostImportModuleDynamically(
     JSContext* aCx, JS::Handle<JS::Value> aReferencingPrivate,
@@ -336,6 +342,12 @@ bool ModuleLoaderBase::HostImportModuleDynamically(
   // Let moduleType be the result of running the module type from module
   // request steps given moduleRequest.
   JS::ModuleType moduleType = JS::GetModuleRequestType(aCx, aModuleRequest);
+  // This check will be moved to HostLoadImportedModule in bug 1820594.
+  if (!ModuleTypeAllowed(moduleType)) {
+    JS_ReportErrorNumberASCII(aCx, js::GetErrorMessage, nullptr,
+                              JSMSG_BAD_MODULE_TYPE);
+    return false;
+  }
 
   // Create a new top-level load request.
   nsCOMPtr<nsIURI> uri = result.unwrap();
@@ -456,7 +468,16 @@ nsresult ModuleLoaderBase::StartOrRestartModuleLoad(ModuleLoadRequest* aRequest,
   MOZ_ASSERT(aRequest->mLoader == this);
   MOZ_ASSERT(aRequest->IsFetching() || aRequest->IsPendingFetchingError());
 
-  aRequest->SetUnknownDataType();
+  // NOTE: The LoadedScript::mDataType field used by the IsStencil call can be
+  //       modified asynchronously after the StartFetch call.
+  //       In order to avoid the race condition, cache the value here.
+  bool isStencil = aRequest->IsStencil();
+
+  MOZ_ASSERT_IF(isStencil, aRestart == RestartRequest::No);
+
+  if (!isStencil) {
+    aRequest->SetUnknownDataType();
+  }
 
   // If we're restarting the request, the module should already be in the
   // "fetching" map.
@@ -472,23 +493,27 @@ nsresult ModuleLoaderBase::StartOrRestartModuleLoad(ModuleLoadRequest* aRequest,
 
   // Check whether the module has been fetched or is currently being fetched,
   // and if so wait for it rather than starting a new fetch.
-  ModuleLoadRequest* request = aRequest->AsModuleRequest();
-
   if (aRestart == RestartRequest::No &&
       ModuleMapContainsURL(
-          ModuleMapKey(request->mURI, aRequest->mModuleType))) {
+          ModuleMapKey(aRequest->mURI, aRequest->mModuleType))) {
     LOG(("ScriptLoadRequest (%p): Waiting for module fetch", aRequest));
-    WaitForModuleFetch(request);
+    WaitForModuleFetch(aRequest);
     return NS_OK;
   }
 
   rv = StartFetch(aRequest);
   NS_ENSURE_SUCCESS(rv, rv);
 
+  if (isStencil) {
+    MOZ_ASSERT(
+        IsModuleFetched(ModuleMapKey(aRequest->mURI, aRequest->mModuleType)));
+    return NS_OK;
+  }
+
   // We successfully started fetching a module so put its URL in the module
   // map and mark it as fetching.
   if (aRestart == RestartRequest::No) {
-    SetModuleFetchStarted(aRequest->AsModuleRequest());
+    SetModuleFetchStarted(aRequest);
   }
 
   return NS_OK;
@@ -707,10 +732,29 @@ nsresult ModuleLoaderBase::CreateModuleScript(ModuleLoadRequest* aRequest) {
     }
 
     MOZ_ASSERT(aRequest->mLoadedScript->IsModuleScript());
-    MOZ_ASSERT(aRequest->mLoadedScript->GetFetchOptions() ==
-               aRequest->mFetchOptions);
-    MOZ_ASSERT(aRequest->mLoadedScript->GetURI() == aRequest->mURI);
-    aRequest->mLoadedScript->SetBaseURL(aRequest->mBaseURL);
+    MOZ_ASSERT(aRequest->mFetchOptions->IsCompatible(
+        aRequest->mLoadedScript->GetFetchOptions()));
+#ifdef DEBUG
+    {
+      bool equals = false;
+      aRequest->mURI->Equals(aRequest->mLoadedScript->GetURI(), &equals);
+      MOZ_ASSERT(equals);
+    }
+#endif
+
+    if (!aRequest->mLoadedScript->BaseURL()) {
+      // If this script is not cached, the BaseURL should be copied from
+      // request to script for later use.
+      aRequest->mLoadedScript->SetBaseURL(aRequest->mBaseURL);
+    } else {
+      // If this script is cached, the BaseURL should match, which is
+      // checked when looking for the cache.
+#ifdef DEBUG
+      bool equals = false;
+      aRequest->mBaseURL->Equals(aRequest->mLoadedScript->BaseURL(), &equals);
+      MOZ_ASSERT(equals);
+#endif
+    }
     RefPtr<ModuleScript> moduleScript =
         aRequest->mLoadedScript->AsModuleScript();
     aRequest->mModuleScript = moduleScript;
@@ -943,10 +987,37 @@ void ModuleLoaderBase::StartFetchingModuleDependencies(
   }
 }
 
+bool ModuleLoaderBase::GetImportMapSRI(
+    nsIURI* aURI, nsIURI* aSourceURI, nsIConsoleReportCollector* aReporter,
+    mozilla::dom::SRIMetadata* aMetadataOut) {
+  MOZ_ASSERT(aMetadataOut->IsEmpty());
+  MOZ_ASSERT(aURI);
+
+  if (!HasImportMapRegistered()) {
+    return false;
+  }
+
+  mozilla::Maybe<nsString> entry =
+      ImportMap::LookupIntegrity(mImportMap.get(), aURI);
+  if (entry.isNothing()) {
+    return false;
+  }
+
+  mozilla::dom::SRICheck::IntegrityMetadata(
+      *entry, aSourceURI->GetSpecOrDefault(), aReporter, aMetadataOut);
+  return true;
+}
+
 void ModuleLoaderBase::StartFetchingModuleAndDependencies(
     ModuleLoadRequest* aParent, const ModuleMapKey& aRequestedModule) {
-  RefPtr<ModuleLoadRequest> childRequest = CreateStaticImport(
-      aRequestedModule.mUri, aRequestedModule.mModuleType, aParent);
+  // Check import map for integrity information
+  mozilla::dom::SRIMetadata sriMetadata;
+  GetImportMapSRI(aRequestedModule.mUri, aParent->mURI,
+                  mLoader->GetConsoleReportCollector(), &sriMetadata);
+
+  RefPtr<ModuleLoadRequest> childRequest =
+      CreateStaticImport(aRequestedModule.mUri, aRequestedModule.mModuleType,
+                         aParent, sriMetadata);
 
   aParent->mImports.AppendElement(childRequest);
 
@@ -990,6 +1061,20 @@ void ModuleLoadRequest::ChildLoadComplete(bool aSuccess) {
     return;
   }
 
+  AutoJSAPI jsapi;
+  if (!jsapi.Init(mLoader->GetGlobalObject())) {
+    return;
+  }
+
+  js::AutoCheckRecursionLimit recursion(jsapi.cx());
+  if (!recursion.check(jsapi.cx())) {
+    // Call LoadFinished on root module directly, as LoadFailed might trigger
+    // another recursive call to sub-modules.
+    mRootModule->SetReady();
+    mRootModule->LoadFinished();
+    return;
+  }
+
   if (!aSuccess) {
     parent->ModuleErrored();
   } else if (parent->mAwaitingImports == 0) {
@@ -1158,9 +1243,15 @@ bool ModuleLoaderBase::HasDynamicImport(
 }
 #endif
 
-JS::Value ModuleLoaderBase::FindFirstParseError(ModuleLoadRequest* aRequest) {
+JS::Value ModuleLoaderBase::FindFirstParseError(JSContext* aCx,
+                                                ModuleLoadRequest* aRequest) {
   MOZ_ASSERT(aRequest);
 
+  js::AutoCheckRecursionLimit recursion(aCx);
+  if (!recursion.check(aCx)) {
+    return JS::UndefinedValue();
+  }
+
   ModuleScript* moduleScript = aRequest->mModuleScript;
   MOZ_ASSERT(moduleScript);
 
@@ -1172,7 +1263,7 @@ JS::Value ModuleLoaderBase::FindFirstParseError(ModuleLoadRequest* aRequest) {
     MOZ_DIAGNOSTIC_ASSERT(moduleScript->HadImportMap() ==
                           childRequest->mModuleScript->HadImportMap());
 
-    JS::Value error = FindFirstParseError(childRequest);
+    JS::Value error = FindFirstParseError(aCx, childRequest);
     if (!error.isUndefined()) {
       return error;
     }
@@ -1195,7 +1286,13 @@ bool ModuleLoaderBase::InstantiateModuleGraph(ModuleLoadRequest* aRequest) {
   ModuleScript* moduleScript = aRequest->mModuleScript;
   MOZ_ASSERT(moduleScript);
 
-  JS::Value parseError = FindFirstParseError(aRequest);
+  AutoJSAPI jsapi;
+  if (NS_WARN_IF(!jsapi.Init(mGlobalObject))) {
+    return false;
+  }
+
+  JSContext* cx = jsapi.cx();
+  JS::Value parseError = FindFirstParseError(cx, aRequest);
   if (!parseError.isUndefined()) {
     moduleScript->SetErrorToRethrow(parseError);
     LOG(("ScriptLoadRequest (%p):   found parse error", aRequest));
@@ -1204,12 +1301,7 @@ bool ModuleLoaderBase::InstantiateModuleGraph(ModuleLoadRequest* aRequest) {
 
   MOZ_ASSERT(moduleScript->ModuleRecord());
 
-  AutoJSAPI jsapi;
-  if (NS_WARN_IF(!jsapi.Init(mGlobalObject))) {
-    return false;
-  }
-
-  JS::Rooted<JSObject*> module(jsapi.cx(), moduleScript->ModuleRecord());
+  JS::Rooted<JSObject*> module(cx, moduleScript->ModuleRecord());
   if (!xpc::Scriptability::AllowedIfExists(module)) {
     return true;
   }
@@ -1319,20 +1411,19 @@ nsresult ModuleLoaderBase::EvaluateModuleInContext(
                             MarkerInnerWindowIdFromJSContext(aCx),
                             profilerLabelString);
 
-  ModuleLoadRequest* request = aRequest->AsModuleRequest();
-  MOZ_ASSERT(request->mModuleScript);
-  MOZ_ASSERT_IF(request->HasScriptLoadContext(),
-                !request->GetScriptLoadContext()->mCompileOrDecodeTask);
+  MOZ_ASSERT(aRequest->mModuleScript);
+  MOZ_ASSERT_IF(aRequest->HasScriptLoadContext(),
+                !aRequest->GetScriptLoadContext()->mCompileOrDecodeTask);
 
-  ModuleScript* moduleScript = request->mModuleScript;
+  ModuleScript* moduleScript = aRequest->mModuleScript;
   if (moduleScript->HasErrorToRethrow()) {
     LOG(("ScriptLoadRequest (%p):   module has error to rethrow", aRequest));
     JS::Rooted<JS::Value> error(aCx, moduleScript->ErrorToRethrow());
     JS_SetPendingException(aCx, error);
     // For a dynamic import, the promise is rejected.  Otherwise an error
     // is either reported by AutoEntryScript.
-    if (request->IsDynamicImport()) {
-      FinishDynamicImport(aCx, request, NS_OK, nullptr);
+    if (aRequest->IsDynamicImport()) {
+      FinishDynamicImport(aCx, aRequest, NS_OK, nullptr);
     }
     return NS_OK;
   }
@@ -1345,16 +1436,16 @@ nsresult ModuleLoaderBase::EvaluateModuleInContext(
     return NS_OK;
   }
 
-  nsresult rv = InitDebuggerDataForModuleGraph(aCx, request);
+  nsresult rv = InitDebuggerDataForModuleGraph(aCx, aRequest);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  if (request->HasScriptLoadContext()) {
+  if (aRequest->HasScriptLoadContext()) {
     TRACE_FOR_TEST(aRequest, "scriptloader_evaluate_module");
   }
 
   JS::Rooted<JS::Value> rval(aCx);
 
-  mLoader->MaybePrepareModuleForBytecodeEncodingBeforeExecute(aCx, request);
+  mLoader->MaybePrepareModuleForBytecodeEncodingBeforeExecute(aCx, aRequest);
 
   bool ok = JS::ModuleEvaluate(aCx, module, &rval);
 
@@ -1362,7 +1453,7 @@ nsresult ModuleLoaderBase::EvaluateModuleInContext(
   // unless the user cancels execution.
   MOZ_ASSERT_IF(ok, !JS_IsExceptionPending(aCx));
 
-  if (!ok || IsModuleEvaluationAborted(request)) {
+  if (!ok || IsModuleEvaluationAborted(aRequest)) {
     LOG(("ScriptLoadRequest (%p):   evaluation failed", aRequest));
     // For a dynamic import, the promise is rejected. Otherwise an error is
     // reported by AutoEntryScript.
@@ -1377,11 +1468,11 @@ nsresult ModuleLoaderBase::EvaluateModuleInContext(
     evaluationPromise.set(&rval.toObject());
   }
 
-  if (request->IsDynamicImport()) {
+  if (aRequest->IsDynamicImport()) {
     if (NS_FAILED(rv)) {
-      FinishDynamicImportAndReject(request, rv);
+      FinishDynamicImportAndReject(aRequest, rv);
     } else {
-      FinishDynamicImport(aCx, request, NS_OK, evaluationPromise);
+      FinishDynamicImport(aCx, aRequest, NS_OK, evaluationPromise);
     }
   } else {
     // If this is not a dynamic import, and if the promise is rejected,
@@ -1394,7 +1485,7 @@ nsresult ModuleLoaderBase::EvaluateModuleInContext(
     }
   }
 
-  rv = mLoader->MaybePrepareModuleForBytecodeEncodingAfterExecute(request,
+  rv = mLoader->MaybePrepareModuleForBytecodeEncodingAfterExecute(aRequest,
                                                                   NS_OK);
 
   mLoader->MaybeTriggerBytecodeEncoding();
