@@ -21,6 +21,7 @@
 #include "jstypes.h"
 
 #include "builtin/Array.h"
+#include "builtin/MapObject.h"
 #include "builtin/SelfHostingDefines.h"
 #include "ds/Sort.h"
 #include "gc/GC.h"
@@ -38,12 +39,6 @@
 #include "vm/Shape.h"
 #include "vm/StringType.h"
 #include "vm/TypedArrayObject.h"
-
-#ifdef ENABLE_RECORD_TUPLE
-#  include "builtin/RecordObject.h"
-#  include "builtin/TupleObject.h"
-#endif
-
 #include "vm/NativeObject-inl.h"
 #include "vm/PlainObject-inl.h"  // js::PlainObject::createWithTemplate
 
@@ -340,43 +335,6 @@ bool PropertyEnumerator::enumerateNativeProperties(JSContext* cx) {
         }
       }
     }
-#ifdef ENABLE_RECORD_TUPLE
-    else {
-      Rooted<RecordType*> rec(cx);
-      if (RecordObject::maybeUnbox(pobj, &rec)) {
-        Rooted<ArrayObject*> keys(cx, rec->keys());
-
-        for (size_t i = 0; i < keys->length(); i++) {
-          JSAtom* key = &keys->getDenseElement(i).toString()->asAtom();
-          PropertyKey id = AtomToId(key);
-          if (!enumerate<CheckForDuplicates>(cx, id,
-                                             /* enumerable = */ true)) {
-            return false;
-          }
-        }
-
-        return true;
-      } else {
-        mozilla::Maybe<TupleType&> tup = TupleObject::maybeUnbox(pobj);
-        if (tup) {
-          uint32_t len = tup->length();
-
-          for (size_t i = 0; i < len; i++) {
-            // We expect tuple indices not to get so large that `i` won't
-            // fit into an `int32_t`.
-            MOZ_ASSERT(PropertyKey::fitsInInt(i));
-            PropertyKey id = PropertyKey::Int(i);
-            if (!enumerate<CheckForDuplicates>(cx, id,
-                                               /* enumerable = */ true)) {
-              return false;
-            }
-          }
-
-          return true;
-        }
-      }
-    }
-#endif
 
     // The code below enumerates shape properties (including sparse elements) so
     // if we can ignore those we're done.
@@ -1625,6 +1583,50 @@ RegExpStringIteratorObject* js::NewRegExpStringIterator(JSContext* cx) {
   return NewObjectWithGivenProto<RegExpStringIteratorObject>(cx, proto);
 }
 
+#ifdef NIGHTLY_BUILD
+static const JSClass IteratorRangePrototypeClass = {
+    "Numeric Range Iterator",
+    0,
+};
+
+enum {
+  IteratorRangeSlotStart,
+  IteratorRangeSlotEnd,
+  IteratorRangeSlotStep,
+  IteratorRangeSlotInclusiveEnd,
+  IteratorRangeSlotZero,
+  IteratorRangeSlotOne,
+  IteratorRangeSlotCurrentCount,
+  IteratorRangeSlotCount
+};
+
+// slot numbers must match constants used in self-hosted code
+static_assert(IteratorRangeSlotStart == ITERATOR_RANGE_SLOT_START);
+static_assert(IteratorRangeSlotEnd == ITERATOR_RANGE_SLOT_END);
+static_assert(IteratorRangeSlotStep == ITERATOR_RANGE_SLOT_STEP);
+static_assert(IteratorRangeSlotInclusiveEnd ==
+              ITERATOR_RANGE_SLOT_INCLUSIVE_END);
+static_assert(IteratorRangeSlotZero == ITERATOR_RANGE_SLOT_ZERO);
+static_assert(IteratorRangeSlotOne == ITERATOR_RANGE_SLOT_ONE);
+static_assert(IteratorRangeSlotCurrentCount ==
+              ITERATOR_RANGE_SLOT_CURRENT_COUNT);
+
+static const JSFunctionSpec iterator_range_methods[] = {
+    JS_SELF_HOSTED_FN("next", "IteratorRangeNext", 0, 0),
+    JS_FS_END,
+};
+
+IteratorRangeObject* js::NewIteratorRange(JSContext* cx) {
+  RootedObject proto(
+      cx, GlobalObject::getOrCreateIteratorRangePrototype(cx, cx->global()));
+  if (!proto) {
+    return nullptr;
+  }
+
+  return NewObjectWithGivenProto<IteratorRangeObject>(cx, proto);
+}
+#endif
+
 // static
 PropertyIteratorObject* GlobalObject::getOrCreateEmptyIterator(JSContext* cx) {
   if (!cx->global()->data().emptyIterator) {
@@ -2144,6 +2146,19 @@ JSObject* GlobalObject::getOrCreateRegExpStringIteratorPrototype(
                               regexp_string_iterator_methods>);
 }
 
+#ifdef NIGHTLY_BUILD
+/* static */
+JSObject* GlobalObject::getOrCreateIteratorRangePrototype(
+    JSContext* cx, Handle<GlobalObject*> global) {
+  return getOrCreateBuiltinProto(
+      cx, global, ProtoKind::IteratorRangeProto,
+      cx->names().RegExp_String_Iterator_.toHandle(),
+      initObjectIteratorProto<ProtoKind::IteratorRangeProto,
+                              &IteratorRangePrototypeClass,
+                              iterator_range_methods>);
+}
+#endif
+
 // Iterator Helper Proposal 2.1.3.1 Iterator()
 // https://tc39.es/proposal-iterator-helpers/#sec-iterator as of revision
 // ed6e15a
@@ -2262,6 +2277,13 @@ const JSClass IteratorHelperObject::class_ = {
     JSCLASS_HAS_RESERVED_SLOTS(IteratorHelperObject::SlotCount),
 };
 
+#ifdef NIGHTLY_BUILD
+const JSClass IteratorRangeObject::class_ = {
+    "IteratorRange",
+    JSCLASS_HAS_RESERVED_SLOTS(IteratorRangeSlotCount),
+};
+#endif
+
 /* static */
 NativeObject* GlobalObject::getOrCreateIteratorHelperPrototype(
     JSContext* cx, Handle<GlobalObject*> global) {
@@ -2310,3 +2332,105 @@ bool js::IterableToArray(JSContext* cx, HandleValue iterable,
   }
   return true;
 }
+
+bool js::HasOptimizableArrayIteratorPrototype(JSContext* cx) {
+  // Return true if %ArrayIteratorPrototype% does not have (re)defined `next`
+  // and `return` properties.
+  return cx->realm()->realmFuses.optimizeArrayIteratorPrototypeFuse.intact();
+}
+
+template <MustBePacked Packed>
+bool js::IsArrayWithDefaultIterator(JSObject* obj, JSContext* cx) {
+  if constexpr (Packed == MustBePacked::Yes) {
+    if (!IsPackedArray(obj)) {
+      return false;
+    }
+  } else {
+    if (!obj->is<ArrayObject>()) {
+      return false;
+    }
+  }
+  ArrayObject* arr = &obj->as<ArrayObject>();
+
+  // Ensure Array.prototype[@@iterator] and %ArrayIteratorPrototype% haven't
+  // been mutated in a way that affects the iterator protocol.
+  if (!arr->realm()->realmFuses.optimizeGetIteratorFuse.intact()) {
+    return false;
+  }
+
+  // Ensure the array has Array.prototype as prototype and doesn't have an own
+  // @@iterator property.
+  //
+  // Most arrays have the default array shape so we have a fast path for this
+  // case.
+  GlobalObject& global = arr->global();
+  if (arr->shape() == global.maybeArrayShapeWithDefaultProto()) {
+    return true;
+  }
+
+  NativeObject* arrayProto = global.maybeGetArrayPrototype();
+  if (!arrayProto || arr->staticPrototype() != arrayProto) {
+    return false;
+  }
+  if (arr->containsPure(PropertyKey::Symbol(cx->wellKnownSymbols().iterator))) {
+    return false;
+  }
+
+  return true;
+}
+
+template bool js::IsArrayWithDefaultIterator<MustBePacked::No>(JSObject* obj,
+                                                               JSContext* cx);
+template bool js::IsArrayWithDefaultIterator<MustBePacked::Yes>(JSObject* obj,
+                                                                JSContext* cx);
+
+template <typename ObjectT, JSProtoKey ProtoKey>
+static bool IsMapOrSetObjectWithDefaultIterator(JSObject* objArg,
+                                                JSContext* cx) {
+  if (!objArg->is<ObjectT>()) {
+    return false;
+  }
+  auto* obj = &objArg->as<ObjectT>();
+
+  // Ensure {Map,Set}.prototype[@@iterator] and %{Map,Set}IteratorPrototype%
+  // haven't been mutated in a way that affects the iterator protocol.
+  //
+  // Note: this doesn't guard against `return` properties on the iterator
+  // prototype, so this should only be used in places where we don't have to
+  // call `IteratorClose`.
+  if constexpr (std::is_same_v<ObjectT, MapObject>) {
+    if (!obj->realm()->realmFuses.optimizeMapObjectIteratorFuse.intact()) {
+      return false;
+    }
+  } else {
+    static_assert(std::is_same_v<ObjectT, SetObject>);
+    if (!obj->realm()->realmFuses.optimizeSetObjectIteratorFuse.intact()) {
+      return false;
+    }
+  }
+
+  // Ensure the object has the builtin prototype as proto.
+  GlobalObject& global = obj->global();
+  JSObject* proto = global.maybeGetPrototype(ProtoKey);
+  if (!proto || obj->staticPrototype() != proto) {
+    return false;
+  }
+
+  // Ensure the Map or Set object doesn't have an own @@iterator property.
+  // Most Maps and Sets have no own properties so we have a fast path for this
+  // case.
+  if (obj->empty()) {
+    return true;
+  }
+  if (obj->containsPure(PropertyKey::Symbol(cx->wellKnownSymbols().iterator))) {
+    return false;
+  }
+  return true;
+}
+bool js::IsMapObjectWithDefaultIterator(JSObject* obj, JSContext* cx) {
+  return IsMapOrSetObjectWithDefaultIterator<MapObject, JSProto_Map>(obj, cx);
+}
+
+bool js::IsSetObjectWithDefaultIterator(JSObject* obj, JSContext* cx) {
+  return IsMapOrSetObjectWithDefaultIterator<SetObject, JSProto_Set>(obj, cx);
+}
